import argparse
from .. import GoodConf


def argparser_add_argument(parser: argparse.ArgumentParser, config: GoodConf):
    """Adds argument for config to existing argparser"""
    help = "Config file."
    if config.file_env_var:
        help += (" Can also be configured via the "
                 "environment variable: {}".format(config.file_env_var))
    if config.default_files:
        help += (" Defaults to the first file that exists from "
                 "[{}].".format(', '.join(config.default_files)))
    parser.add_argument('-C', '--config', metavar='FILE', help=help)



import os
from typing import TypeVar, Callable
from decimal import Decimal

from distutils.util import strtobool

CASTS = [int, str, float, list, bool, Decimal]
CastTypes = TypeVar('CastTypes', *CASTS)


class RequiredValueMissing(Exception):
    pass


class Value:
    def __init__(self, key: str=None, default=None,
                 initial: Callable[[], CastTypes] = None,
                 cast_as: CastTypes = None, help: str = ""):
        """
        :param key:      Name of the value used in file or environment
                         variable. Set automatically by the GoodConf metaclass.
        :param default:  Default value if none is provided. If left unset,
                         loading a config thait fails to provide this value
                         will raise a RequiredValueMissing exception.
        :param initial:  Initial value to use when generating a config
        :param cast_as:  Python type to cast variable as. Defaults to type of
                         default (if provided) or str.
        :param help:     Plain-text description of the value.
        """
        self.key = key
        self.default = default
        self.initial = initial
        self.help = help
        if cast_as:
            self.cast_as = cast_as
        elif default is not None:
            self.cast_as = type(default)
        else:
            self.cast_as = str

    @property
    def required(self):
        return self.default is None

    @property
    def value(self):
        if self.key in os.environ:
            return self.cast(os.environ[self.key])
        if not hasattr(self, '_value'):
            self._value = self.default
            if callable(self._value):
                self._value = self._value()
        return self._value

    @value.setter
    def value(self, value):
        if value is None:
            if self.required:
                raise RequiredValueMissing(self.key)
            del self.value
        else:
            self._value = value

    @value.deleter
    def value(self):
        if hasattr(self, '_value'):
            del self._value

    @property
    def initial(self):
        return self._initial() if self._initial else self.default or ''

    @initial.setter
    def initial(self, value):
        if value and not callable(value):
            raise ValueError("Initial value must be a callable.")
        self._initial = value

    def __get__(self, instance, owner):
        if instance is None:
            # Accessing via the class. Return the initial value (falling
            # back to default).
            return self.initial
        return self.value

    def __set__(self, instance, value):
        self.value = value

    def __delete__(self, instance):
        del self.value

    def cast(self, val: str):
        """converts string to type requested by `cast_as`"""
        try:
            return getattr(self, 'cast_as_{}'.format(
                self.cast_as.__name__.lower()))(val)
        except AttributeError:
            return self.cast_as(val)

    def cast_as_list(self, val: str) -> list:
        """Convert a comma-separated string to a list"""
        return val.split(',')

    def cast_as_bool(self, val: str) -> bool:
        """
        True values are y, yes, t, true, on and 1
        False values are n, no, f, false, off and 0
        Raises ValueError if val is anything else.
        """
        return bool(strtobool(val))



import argparse
from typing import List
from contextlib import contextmanager

from goodconf import GoodConf
from .argparse import argparser_add_argument


@contextmanager
def load_config_from_cli(config: GoodConf, argv: List[str]) -> List[str]:
    """Loads config, checking CLI arguments for a config file"""

    # Monkey patch Django's command parser
    from django.core.management.base import BaseCommand
    original_parser = BaseCommand.create_parser

    def patched_parser(self, prog_name, subcommand):
        parser = original_parser(self, prog_name, subcommand)
        argparser_add_argument(parser, config)
        return parser

    BaseCommand.create_parser = patched_parser

    try:
        parser = argparse.ArgumentParser(add_help=False)
        argparser_add_argument(parser, config)

        config_arg, default_args = parser.parse_known_args(argv)
        config.load(config_arg.config)
        yield default_args
    finally:
        # Put that create_parser back where it came from or so help me!
        BaseCommand.create_parser = original_parser


def execute_from_command_line_with_config(config: GoodConf, argv: List[str]):
    """Load's config then runs Django's execute_from_command_line"""
    with load_config_from_cli(config, argv) as args:
        from django.core.management import execute_from_command_line
        execute_from_command_line(args)
